import { Injectable } from '@nestjs/common';
import { EntityManager } from '@mikro-orm/core';
import { InjectRepository } from '@mikro-orm/nestjs';
import { EntityRepository } from '@mikro-orm/mysql';
import { Activity } from '../activities/activities.entity';
import { SessionPoints } from '../points/sessionPoints.entity';
import { RoomSession } from '../room/roomSession.entity';
import { RoomSessionUser } from '../room/roomSessionUser.entity';
import { ActivitiesService } from '../activities/activities.service';

export enum ActivityType {
  NOTE_CREATE = 'note.create',
  NOTE_UPDATE = 'note.update',
  WHITEBOARD = 'whiteboard',
  VIDEO_ON = 'video.on',
  VIDEO_OFF = 'video.off',
  AUDIO_ON = 'audio.on',
  AUDIO_OFF = 'audio.off',
  AUDIO_TALKING = 'audio.talking',
  AUDIO_NOT_TALKING = 'audio.nottalking',
  BROWSER_OPEN = 'browser.open',
  BROWSER_CLOSE = 'browser.close',
  BROWSER_MOUSE = 'browser.mouse',
  BROWSER_KEY = 'browser.key',
  WHITEBOARD_ERASER_USE = 'whiteboard.eraser',
  NOTE_DELETE = 'note.delete',
  NOTE_DELETE_CHARS = 'note.delete.chars',
}

/**
 * A service responsible for managing to track the activities.
 */
@Injectable()
export class SessionPointsService {
  /**
    * Initializes the ActivitiesService.
    * @param em - The EntityManager instance provided by MikroORM.
    * @param repository - The repository for the Activity entity.
    */
  /**
   * Initializes the service with a manager and repository.
   */
  constructor(
    private readonly em: EntityManager,
  ) { }


/**
  /**
   * Calculates and returns points for user activities in a time range.
   */
  public async calculatePoints(
    roomSession: RoomSession,
  ) {
    const activities = await roomSession.activities.loadItems();
    const users = await roomSession.users.loadItems();
    
    const activitiesByUsers = activities.reduce((actsByUsers, activity) => {
      const user = users.find((user) => user.userId === activity.userId);
      let acts = actsByUsers.get(user) ?? [];
      acts = [...acts, activity];
  
      actsByUsers.set(user, acts);
  
      return actsByUsers;
    }, new Map<RoomSessionUser, Activity[]>());

    const pointsArray: SessionPoints[] = [];
  for (const [user, userActivities] of activitiesByUsers) {
    let points = await this.em.findOne(SessionPoints, { roomSessionUser: user });
    if (!points) {
      points = new SessionPoints(user);
    }
    points.audioPoints = await this.calculateAudioPoints(userActivities),
    points.videoPoints = await this.calculateVideoPoints(userActivities),
    points.browserPoints = await this.calculateBrowserPoints(userActivities),
    points.whiteboardPoints = await this.calculateWhiteboardPoints(userActivities),
    points.notePoints = await this.calculateNoteAddingPoints(userActivities),
    points.eraserPoints = await this.calculateEraserPoints(userActivities),
    points.noteDeletePoints = await this.calculateNoteDeletePoints(userActivities),

    console.log('Points:', points);

    await this.em.persistAndFlush(points);
    pointsArray.push(points);
  }

  return pointsArray;

  }

  /**
   * Calculates how many audio points (1 point per 5 seconds of talking) a user gets.
   */
  private async calculateAudioPoints(activities: Activity[]) {
    const audioActivities = activities
      .filter(a => a.activity.startsWith('audio.'))
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    let totalPoints = 0;
    let talkingStart: Date | null = null;

    for (const act of audioActivities) {
      if (act.activity === 'audio.talking' && !talkingStart) {
        talkingStart = act.timestamp;
      } else if ((act.activity === 'audio.nottalking' || act.activity === 'audio.off') && talkingStart) {
        const talkingDuration = (act.timestamp.getTime() - talkingStart.getTime()) / 1000; // in seconds
        if (talkingDuration <= 2) {
          totalPoints += 1; // Short sentence
        } else if (talkingDuration <= 5) {
          totalPoints += 2; // Middle sentence
        } else {
          totalPoints += 4; // Long sentence
        }
        talkingStart = null;
      }
    }

    return totalPoints;
  }

  /**
   * Calculates how many video points (1 point per 5 minutes of video on) a user gets.
   */
  private async calculateVideoPoints(activities: Activity[]) {
    const videoActivities = activities
      .filter(a => a.activity.startsWith('video.'))
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    let totalOnTime = 0;
    let videoStart: Date | null = null;

    for (const act of videoActivities) {
      if (act.activity === 'video.on' && !videoStart) {
        videoStart = act.timestamp;
      } else if (act.activity === 'video.off' && videoStart) {
        totalOnTime += act.timestamp.getTime() - videoStart.getTime();
        videoStart = null;
      }
    }

    const videoPoints = Math.floor(totalOnTime / 300000);
    return videoPoints;
  }

  /**
   * Calculates how many browser points (1 point per 50 key events, 1 point per 10 mouse events).
   */
  private async calculateBrowserPoints(activities: Activity[]) {
    const browserActivities = activities
      .filter(a => a.activity.startsWith('browser.'))
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    const keyCount = browserActivities.filter(a => a.activity === 'browser.key').length;
    const mouseCount = browserActivities.filter(a => a.activity === 'browser.mouse').length;

    const keyPoints = Math.floor(keyCount / 50);
    const mousePoints = Math.floor(mouseCount / 10);

    return keyPoints + mousePoints;
  }

  /**
   * Calculates how many whiteboard points (1 point per 5 whiteboard events) a user gets.
   */
  private async calculateWhiteboardPoints(activities: Activity[]) {
    const whiteboardActivities = activities
      .filter(a => a.activity === 'whiteboard')
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    const whiteboardPoints = Math.floor(whiteboardActivities.length / 5);

    return whiteboardPoints;
  }

  private async calculateEraserPoints(activities: Activity[]) {
    const eraserActivities = activities
      .filter(a => a.activity === 'whiteboard.eraser.use');

    const eraserPoints = eraserActivities.length;

    return eraserPoints;
  }
  /**
   * Calculates how many note points (1 point for creating a note, 1 point per 50 characters added).
   */
  private async calculateNoteAddingPoints(activities: Activity[]) {
    const noteActivities = activities
      .filter(a => a.activity.startsWith('note.'))
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    let noteAddingPoints = 0;
    for (const act of noteActivities) {
      if (act.activity === 'note.create') {
        noteAddingPoints++;
      } else if (act.activity === 'note.update' && act.numericValue) {
        noteAddingPoints += Math.floor(act.numericValue / 50);
      } else if (act.activity === 'note.update' && act.numericValue < 0) {
        noteAddingPoints += Math.ceil(act.numericValue / 50);
      }
    }
    return noteAddingPoints;
  }

  private async calculateNoteDeletePoints(activities: Activity[]) {
    const noteActivities = activities
      .filter(a => a.activity === 'note.delete' || a.activity === 'note.delete.chars')
      .sort((a, b) => a.timestamp.getTime() - b.timestamp.getTime());

    let noteDeletionPoints = 0;

    for (const act of noteActivities) {
      if (act.activity === 'note.delete') {
        noteDeletionPoints += 1; // 1 point for deleting a note
      } else if (act.activity === 'note.delete.chars' && act.numericValue) {
        noteDeletionPoints += Math.floor(act.numericValue / 50); // 1 point for every 50 characters deleted
      }
    }

    return noteDeletionPoints;
  }

}
