import { Injectable, forwardRef, Inject } from '@nestjs/common';
import { EntityManager } from '@mikro-orm/core';
import { InjectRepository } from '@mikro-orm/nestjs';
import { EntityRepository } from '@mikro-orm/mysql';
import { RoomSession } from './roomSession.entity';
import { Room } from './room.entity'; 
import { RoomSessionUser } from './roomSessionUser.entity';
import { SessionPoints } from '../points/sessionPoints.entity';
import { RoomSessionUserService } from './roomSessionUser.service';
import { RoomSessionPointsRequestDto } from './roomSession.dto';
import { ActivitiesService } from 'src/activities/activities.service';

export enum ModeType {
    PAIR_PROGRAMMING = 'pairProgramming',
    ROLE_BASED_PROGRAMMING = 'roleBasedProgramming',
    DEFAULT = 'default',
}

/**
 * A service responsible for managing room sessions.
 */
@Injectable()
export class RoomSessionService {
    /**
     * Initializes the service with a manager and repository.
     * @param em - The EntityManager instance provided by MikroORM.
     * @param repository - The repository for the RoomSession entity.
     */
    constructor(
        private readonly em: EntityManager,
        @InjectRepository(RoomSession)
        private readonly repository: EntityRepository<RoomSession>,
        @InjectRepository(RoomSessionUser)
        private readonly roomSessionUserRepository: EntityRepository<RoomSessionUser>,
        @Inject(forwardRef(() => RoomSessionUserService))
        private readonly roomSessionUserService: RoomSessionUserService,
        private readonly activitiesService: ActivitiesService,
    ) {}

    /**
     * Begins a new session for the specified room.
     * 
     * This method creates a new `RoomSession` instance with the provided room,
     * the current date as the start time, and a null end time. It then persists
     * the session to the database and returns the created session.
     * 
     * @param room - The room for which the session is being started.
     * @returns A promise that resolves to the newly created `RoomSession` instance.
     */
    public async beginSession(room: Room): Promise<RoomSession> {
        const roomSession = new RoomSession(room);
        await this.em.persistAndFlush(roomSession);
        return roomSession;
    }

    /**
     * Closes the given room session by setting the `dateClosed` property to the current date and time.
     * Persists the changes to the database by flushing the entity manager.
     *
     * @param roomSession - The room session to be closed.
     * @returns A promise that resolves when the session is successfully closed.
     */
    public async closeSession(roomSession: RoomSession): Promise<void> {
        roomSession.dateClosed = new Date();
        await this.em.persistAndFlush(roomSession);
    }

    public async getRoomSessionById(id: number) {
        try {
            const roomSession = await this.repository.findOneOrFail({ id });
            const users = await this.em.find(RoomSessionUser, { roomSession: roomSession.id });
            return {
                id: roomSession.id,
                dateOpened: roomSession.dateOpened,
                dateClosed: roomSession.dateClosed,
                roomId: roomSession.room.id,
                users: users.map(user => ({
                    id: user.id,
                    userId: user.userId,
                    name: user.name
                }))
            };
        }
        catch (error) {
            console.error('Error in getRoomSessionById:', error);
            throw error;
        }
    }
    

    public async getallPositivePointsBySession(sessionId: number): Promise<RoomSessionPointsRequestDto> {
        const roomSessionUsers = await this.roomSessionUserRepository.find({ roomSession: sessionId });
        const points = await Promise.all(roomSessionUsers.map(user => this.roomSessionUserService.getPointsByRoomSessionUser(user.id)));
        const totalPoints = points.reduce((total, userPoints) => total + userPoints.totalPoints, 0);
        const allPositivePoints = points.reduce((acc, userPoints) => {
            acc.audioPoints += userPoints.allPositivePoints.audioPoints;
            acc.videoPoints += userPoints.allPositivePoints.videoPoints;
            acc.notePoints += userPoints.allPositivePoints.notePoints;
            acc.whiteboardPoints += userPoints.allPositivePoints.whiteboardPoints;
            acc.browserPoints += userPoints.allPositivePoints.browserPoints;
            return acc;
        }, {
            audioPoints: 0,
            videoPoints: 0,
            notePoints: 0,
            whiteboardPoints: 0,
            browserPoints: 0,
        });
        const allNegativePoints = points.reduce((acc, userPoints) => {
            acc.eraserPoints += userPoints.allNegativePoints.eraserPoints;
            acc.noteDeletePoints += userPoints.allNegativePoints.noteDeletePoints;
            return acc;
        }
        , {
            eraserPoints: 0,
            noteDeletePoints: 0,
        });
        

        return { allPositivePoints, totalPoints, sessionId, allNegativePoints };
    }

    async createMode(roomSession: RoomSession, modeType: ModeType): Promise<void> {
        const modeName = modeType.toString();
        roomSession.modeName = modeName;
        await this.em.persistAndFlush(roomSession);
    }

    public async assignMode(sessionId: number): Promise<ModeType> {
        const roomSession = await this.repository.findOneOrFail({ id: sessionId });
        let modeType = ModeType.DEFAULT;
        if (this.shouldAssignPairProgramming(roomSession)) {
            modeType = ModeType.PAIR_PROGRAMMING;
        }
        else if (this.shouldAssignRoleBasedProgramming(roomSession)) {
            modeType = ModeType.ROLE_BASED_PROGRAMMING;
        }
        this.createMode(roomSession, modeType);
        return modeType;
    }

    private async shouldAssignPairProgramming(roomSession: RoomSession): Promise<boolean> {
        const users = await roomSession.users.loadItems();
        return users.length === 2;
    }

    private async shouldAssignRoleBasedProgramming(roomSession: RoomSession): Promise<boolean> {
            const users = await roomSession.users.loadItems();
            return users.length > 2;
        }

}
