import { Injectable, Inject, forwardRef } from '@nestjs/common';
import { EntityManager } from '@mikro-orm/core';
import { InjectRepository } from '@mikro-orm/nestjs';
import { EntityRepository } from '@mikro-orm/mysql';
import { RoomSessionUser } from './roomSessionUser.entity';
import { RoomSession } from './roomSession.entity';
import { RoomSessionUserPointsResponseDto } from './roomSessionUser.dto';
import { SessionPoints } from '../points/sessionPoints.entity';
import { ActivitiesService } from '../activities/activities.service';
import { RoomSessionPointsRequestDto } from './roomSession.dto';
import { RoomSessionService } from './roomSession.service';

export enum RoleType {
    COORDINATOR = 'COORDINATOR',
    DOCUMENTER  = 'DOCUMENTER',
    CODER_DEBUGGER  = 'CODER_DEBUGGER',
    VISUALIZER  = 'VISUALIZER',
    ACTIVE_PARTICIPANT  = 'ACTIVE_PARTICIPANT',
    OBSERVER  = 'OBSERVER',
    LURKER  = 'LURKER',
    DISRUPTOR   = 'DISRUPTOR',
}
/**
 * A service responsible for managing room sessions and users.
 */
@Injectable()
export class RoomSessionUserService {
    /**
     * Initializes the service with a manager and repository.
     * @param em - The EntityManager instance provided by MikroORM.
     * @param repository - The repository for the Room entity.
     */
    constructor(
        private readonly em: EntityManager,
        @InjectRepository(SessionPoints)
        private readonly pointsRepository: EntityRepository<SessionPoints>,
        @InjectRepository(RoomSessionUser)
        private readonly userRepository: EntityRepository<RoomSessionUser>,
        private readonly activitiesService: ActivitiesService,
        @Inject(forwardRef(() => RoomSessionService))
        private readonly roomSessionService: RoomSessionService,
    ) {}

    /**
     * Adds a new user to a room session.
     *
     * @param roomSession - The room session to which the user will be added.
     * @param userId - The unique identifier of the user to be added.
     * @param name - The name of the user to be added.
     * @returns A promise that resolves to the newly created `RoomSessionUser` instance.
     */
    public async addSessionUser(roomSession: RoomSession, userId: string, name: string): Promise<RoomSessionUser> {
        const roomSessionUser = new RoomSessionUser(roomSession, userId, name);
        await this.em.persistAndFlush(roomSessionUser);
        return roomSessionUser;
    }
    
    /**
     * Updates the name of a user in a room session. If the user already exists in the session,
     * their name is updated. If the user does not exist, a new user is created and added to the session.
     *
     * @param roomSession - The room session to which the user belongs.
     * @param userId - The unique identifier of the user.
     * @param name - The new name to assign to the user.
     * @returns The updated or newly created RoomSessionUser instance.
     * @throws An error if the operation fails during persistence.
     */
    async updateSessionUserName(roomSession: RoomSession, userId: string, name:string) {
        let user = await this.userRepository.findOne({ roomSession, userId });
        if (user) {
            user.name = name;
        } else {
            user = new RoomSessionUser(roomSession, userId, name);
        }
        await this.em.persistAndFlush(user);
        return user;
    }

    /**
        * Retrieves detailed points breakdown by activity type for a specific user.
        * 
        * @param userId - The unique identifier of the user whose points are to be retrieved.
        * @returns A promise that resolves to an object containing the detailed points breakdown.
        */
        public async getPointsByRoomSessionUser(userId: number): Promise<RoomSessionUserPointsResponseDto> {
            const pointsEntries = await this.pointsRepository.find({ roomSessionUser: userId });
        
            let totalPoints = 0;
            let audioPoints = 0;
            let videoPoints = 0;
            let browserPoints = 0;
            let whiteboardPoints = 0;
            let notePoints = 0;
            let eraserPoints = 0;
            let noteDeletePoints = 0;
        
            pointsEntries.forEach(entry => {
                audioPoints += entry.audioPoints;
                videoPoints += entry.videoPoints;
                browserPoints += entry.browserPoints;
                whiteboardPoints += entry.whiteboardPoints;
                eraserPoints += entry.eraserPoints;
                notePoints += entry.notePoints;
                noteDeletePoints += entry.noteDeletePoints;
                totalPoints += entry.audioPoints + entry.videoPoints + entry.browserPoints + 
                          entry.whiteboardPoints + entry.notePoints;
        });        
            return {
                totalPoints,
                userId,
                allPositivePoints: {
                    audioPoints,
                    videoPoints,
                    browserPoints,
                    whiteboardPoints,
                    notePoints,
                },
                allNegativePoints: {
                    eraserPoints,
                    noteDeletePoints,
                },
            };
        }

        public async createRole(roomSessionUser: RoomSessionUser, roleType: RoleType): Promise<void> {
            // Initialize roleName as an array if it doesn't exist
            if (!roomSessionUser.roleName) {
                roomSessionUser.roleName = [];
            }
            
            // Add role if it doesn't already exist
            if (!roomSessionUser.roleName.includes(roleType)) {
                roomSessionUser.roleName.push(roleType);
                await this.em.persistAndFlush(roomSessionUser);
            }
        }

        /**
         * Assigns roles to a given RoomSessionUser based on their points and the session's total points.
         * 
         * This method evaluates the user's points in the context of the session and determines
         * which roles should be assigned to them. The roles are then created and associated
         * with the user in the system.
         * 
         * @param roomSessionUser - The RoomSessionUser entity for which roles are to be assigned.
         * @returns A promise that resolves to an array of assigned RoleType values.
         */
        public async assignRole(userId: number): Promise<RoleType[]> {
            const roomSessionUser = await this.userRepository.findOneOrFail({ id: userId });
            const userPoints = await this.getPointsByRoomSessionUser(roomSessionUser.id);
            const sessionPoints = await this.roomSessionService.getallPositivePointsBySession(roomSessionUser.roomSession.id);

            let roleTypes: RoleType[] = [];
            if (this.shouldAssignCoordinator(userPoints, sessionPoints)) {
                roleTypes.push(RoleType.COORDINATOR);
            } 
            if (await this.shouldAssignDocumenter(userPoints, sessionPoints)) {
                roleTypes.push(RoleType.DOCUMENTER);
            }
            if (await this.shouldAssignLurker(userPoints)) {
                roleTypes.push(RoleType.LURKER);
            }
            if (await this.shouldAssignDisrupter(userPoints)) {
                roleTypes.push(RoleType.DISRUPTOR);
            }
            if (this.shouldAssignActiveParticipant(userPoints)) {
                roleTypes.push(RoleType.ACTIVE_PARTICIPANT);
            }
            if (this.shouldAssignObserver(userPoints)) {
                roleTypes.push(RoleType.OBSERVER);
            }
            if (this.shouldAssignCoder(userPoints, sessionPoints)) {
                roleTypes.push(RoleType.CODER_DEBUGGER);
            }
            if (this.shouldAssignVisualizer(userPoints, sessionPoints)) {
                roleTypes.push(RoleType.VISUALIZER);
            }
            for (let role of roleTypes) {
                await this.createRole(roomSessionUser, role);
            }
            console.log('roleTypes', roleTypes);
            return roleTypes;
        }

        
        /**
         * Determines whether a user should be assigned as a coordinator based on their audio points contribution.
         *
         * This method compares the ratio of the user's audio points to the total session audio points.
         * If the user's contribution exceeds 50% of the total session audio points, the method returns `true`,
         * indicating that the user should be assigned as a coordinator. Otherwise, it returns `false`.
         *
         * @param userPoints - An object containing the user's points, including audio points.
         * @param sessionPoints - An object containing the total session points, including audio points.
         * @returns A promise that resolves to `true` if the user should be assigned as a coordinator, or `false` otherwise.
         */
        private shouldAssignCoordinator(userPoints: RoomSessionUserPointsResponseDto, sessionPoints: RoomSessionPointsRequestDto): boolean {
            if (userPoints.allPositivePoints.audioPoints === 0) return false;
            let shouldAssignCoordinator = false;
            if (userPoints.allPositivePoints.audioPoints/sessionPoints.allPositivePoints.audioPoints > 0.5) {
                shouldAssignCoordinator = true;
            }
            return shouldAssignCoordinator;
        }

        
        /**
         * Determines whether a user should be assigned as the documenter in a room session.
         *
         * This method evaluates if a user has contributed significantly to the session's note points
         * compared to the total note points of the session. If the user's note points exceed 50% of the
         * session's total note points, the user is considered for the documenter role. The method also
         * retrieves the total number of users in the session for additional context.
         *
         * @param userPoints - An object containing the user's session ID and their contribution points.
         * @param sessionPoints - An object containing the total points for the session.
         * @returns A promise that resolves to a boolean indicating whether the user should be assigned as the documenter.
         */
        private async shouldAssignDocumenter(userPoints: RoomSessionUserPointsResponseDto, sessionPoints: RoomSessionPointsRequestDto): Promise<boolean> {
            if (sessionPoints.allPositivePoints.notePoints === 0) return false;
            let shouldAssignDocumenter = false;
            const userCount = await this.userRepository.count({ roomSession: userPoints.sessionId });
            if (userPoints.allPositivePoints.notePoints/sessionPoints.allPositivePoints.notePoints > 0.5) {
                shouldAssignDocumenter = true;
            }
            return shouldAssignDocumenter;
        }

        /**
         * Determines whether a user should be assigned as a coder based on their browser points
         * relative to the total session browser points.
         *
         * @param userPoints - An object containing the user's points, including browser points.
         * @param sessionPoints - An object containing the session's total points, including browser points.
         * @returns A promise that resolves to `true` if the user's browser points exceed 50% of the session's total browser points, otherwise `false`.
         */
        public shouldAssignCoder(userPoints: RoomSessionUserPointsResponseDto, sessionPoints: RoomSessionPointsRequestDto): boolean {
            if (sessionPoints.allPositivePoints.browserPoints === 0) return false;
            let shouldAssignCoder = false;
            if (userPoints.allPositivePoints.browserPoints / sessionPoints.allPositivePoints.browserPoints > 0.5) {
                shouldAssignCoder = true;
            }
            return shouldAssignCoder;
        }

        /**
         * Determines whether a user should be assigned as a visualizer based on their contribution
         * to the whiteboard points relative to the total session whiteboard points.
         *
         * @param userPoints - An object containing the user's points, including their whiteboard points.
         * @param sessionPoints - An object containing the total session points, including all whiteboard points.
         * @returns A promise that resolves to a boolean indicating whether the user should be assigned as a visualizer.
         */
        public shouldAssignVisualizer(userPoints: RoomSessionUserPointsResponseDto, sessionPoints: RoomSessionPointsRequestDto): boolean {
            if (sessionPoints.allPositivePoints.whiteboardPoints === 0) return false;
            let shouldAssignVisualizer = false;
            if (userPoints.allPositivePoints.whiteboardPoints / sessionPoints.allPositivePoints.whiteboardPoints > 0.5) {
                shouldAssignVisualizer = true;
            }
            return shouldAssignVisualizer;
        }

        /**
         * Determines whether a user should be assigned as an active participant
         * based on their accumulated points in various activity categories.
         *
         * @param userPoints - An object containing the user's points in different
         * activity categories, including audio, video, browser, whiteboard, and notes.
         * 
         * @returns A promise that resolves to `true` if the user has more than 0 points
         * in all activity categories (audio, video, browser, whiteboard, and notes),
         * otherwise resolves to `false`.
         */
        public shouldAssignActiveParticipant(userPoints: RoomSessionUserPointsResponseDto): boolean {
            return (
            userPoints.allPositivePoints.audioPoints > 0 &&
            userPoints.allPositivePoints.videoPoints > 0 &&
            userPoints.allPositivePoints.browserPoints > 0 &&
            userPoints.allPositivePoints.whiteboardPoints > 0 &&
            userPoints.allPositivePoints.notePoints > 0
            );
        }

        /**
         * Determines whether a user should be assigned as an observer based on their activity points.
         *
         * This method evaluates the user's points across various activity categories (video, audio, browser, notes, and whiteboard)
         * and checks two conditions:
         * 1. The user's video points are the highest among all activity categories.
         * 2. The user's total points are less than 10% of the sum of their points in the other activity categories 
         *    (audio, browser, notes, and whiteboard).
         *
         * If both conditions are met, the method returns `true`, indicating that the user should be assigned as an observer.
         * Otherwise, it returns `false`.
         *
         * @param userPoints - An object containing the user's activity points across different categories and their total points.
         * @returns A promise that resolves to `true` if the user should be assigned as an observer, or `false` otherwise.
         */
        public shouldAssignObserver(userPoints: RoomSessionUserPointsResponseDto): boolean{
            let shouldAssignObserver = false;
            if (
                userPoints.allPositivePoints.videoPoints > userPoints.allPositivePoints.audioPoints &&
                userPoints.allPositivePoints.videoPoints > userPoints.allPositivePoints.browserPoints &&
                userPoints.allPositivePoints.videoPoints > userPoints.allPositivePoints.notePoints &&
                userPoints.allPositivePoints.videoPoints > userPoints.allPositivePoints.whiteboardPoints &&
                userPoints.totalPoints < 0.1 * (userPoints.allPositivePoints.audioPoints + userPoints.allPositivePoints.browserPoints + userPoints.allPositivePoints.notePoints + userPoints.allPositivePoints.whiteboardPoints)
            ) {
                shouldAssignObserver = true;
            }
            return shouldAssignObserver;
        }

        /**
         * Determines whether a user should be assigned the "lurker" role based on their activity points
         * compared to other users in the same room session.
         *
         * This method evaluates if the given user's activity points (audio, video, browser, whiteboard, 
         * and note points) are the lowest among all users in the same room session. If the user has 
         * participated in more than one activity and their points are the lowest across all activities, 
         * they are considered a "lurker."
         *
         * @param userPoints - An object containing the user's activity points and their associated user ID.
         * @returns A promise that resolves to `true` if the user should be assigned the "lurker" role, 
         *          otherwise `false`.
         */
        public async shouldAssignLurker(userPoints: RoomSessionUserPointsResponseDto): Promise<boolean> {
            let shouldAssignLurker = false;
            const allUsersPoints = await this.pointsRepository.find({ roomSessionUser: userPoints.userId });

            if (allUsersPoints.length > 1) {
                const isLowestInAllActivities = allUsersPoints.every(otherUserPoints => 
                    otherUserPoints.roomSessionUser.id === userPoints.userId ||
                    (userPoints.allPositivePoints.audioPoints <= otherUserPoints.audioPoints &&
                     userPoints.allPositivePoints.videoPoints <= otherUserPoints.videoPoints &&
                     userPoints.allPositivePoints.browserPoints <= otherUserPoints.browserPoints &&
                     userPoints.allPositivePoints.whiteboardPoints <= otherUserPoints.whiteboardPoints &&
                     userPoints.allPositivePoints.notePoints <= otherUserPoints.notePoints)
                );

                if (isLowestInAllActivities) {
                    shouldAssignLurker = true;
                }
            }

            return shouldAssignLurker;
        }

        /**
         * Determines whether a user should be assigned the role of "Disrupter" based on their activity points
         * and their contribution to disruptive activities in a session.
         *
         * This method evaluates the user's points in disruptive activities (e.g., note deletions and eraser usage)
         * relative to other users in the same session. It also checks if the user's talking activity (audio points)
         * constitutes at least 20% of the total talking activity in the session. If both conditions are met, the
         * user is assigned the "Disrupter" role.
         *
         * @param userPoints - The points data of the user being evaluated, including their activity points and session ID.
         * @returns A promise that resolves to `true` if the user should be assigned the "Disrupter" role, or `false` otherwise.
         */
        public async shouldAssignDisrupter(userPoints: RoomSessionUserPointsResponseDto): Promise<boolean> {
            if (userPoints.allNegativePoints.eraserPoints === 0 && userPoints.allNegativePoints.noteDeletePoints === 0) return false;
            let shouldAssignDisrupter = false;
            const allUsersPoints = await this.pointsRepository.find({ roomSessionUser: userPoints.userId });
            const sessionPoints = await this.roomSessionService.getallPositivePointsBySession(userPoints.sessionId);

            if (allUsersPoints.length > 1) {
            const isHighestInDisruptiveActivities = allUsersPoints.every(otherUserPoints => 
                otherUserPoints.roomSessionUser.id === userPoints.userId ||
                (userPoints.allNegativePoints.noteDeletePoints >= otherUserPoints.noteDeletePoints &&
                 userPoints.allNegativePoints.eraserPoints >= otherUserPoints.eraserPoints)
            );

            const talkingActivity = userPoints.allPositivePoints.audioPoints / sessionPoints.allPositivePoints.audioPoints >= 0.2;

            if (isHighestInDisruptiveActivities && talkingActivity) {
                shouldAssignDisrupter = true;
            }
            }

            return shouldAssignDisrupter;
        }
    }

